from pyspark.sql import SparkSession
import pandas as pd
import findspark
import pyspark
import numpy as np
from pyspark.sql.types import StructType, StructField, StringType, IntegerType, FloatType
import random as rn
import argparse

def create_example_data_spark(n_persons: int = 1000, n_periods: int = 20
) -> pyspark.sql.DataFrame:
    findspark.init()
    spark = SparkSession.builder.getOrCreate()
    schema = StructType([
        StructField('individual', IntegerType(), True),
        StructField('period', IntegerType(), True),
        StructField('feature_1', FloatType(), True),
        StructField('feature_2', StringType(), True),
        StructField('feature_3', IntegerType(), True),
        StructField('feature_4', StringType(), True)])
    values = spark.createDataFrame([], schema)
    for i in np.arange(n_persons):
            period = np.random.randint(n_periods) + 1
            x_1 = np.random.uniform()
            x_2 = rn.choice(["A", "B", "C"])
            x_3 = np.random.uniform() + 1.0
            x_4_categories = [1, 2, 3, 'a', 'b', 'c', np.nan]
            x_4 = rn.choice(x_4_categories)
            while period <= n_periods:
                values = values.union(spark.createDataFrame([(int(i), period, x_1, x_2, x_3, x_4)]))
                if x_2 == "A":
                    x_1 += np.random.uniform(0, 0.1)
                else:
                    x_1 += np.random.uniform(0, 0.2)
                if x_1 > np.sqrt(x_3):
                    break
                if x_4 in x_4_categories[:-2]:
                    x_4_transition_value = x_4_categories[x_4_categories.index(x_4) + 1]
                    if np.random.uniform() >= 0.75:
                        x_4 = x_4_transition_value
                        del x_4_transition_value
                period += 1
    values = values.withColumn('feature_5', values.feature_2)
    return values

def create_example_data(
    n_persons: int = 8192, n_periods: int = 20, seed_value:int = 9999
) -> pyspark.sql.DataFrame:
    """
    Fabricate an unbalanced panel dataset suitable as FIFE input.

    Args:
        n_persons: the number of people to be in the dataset
        n_periods: the number of periods to be in the dataset
        seed_value: seed for random value generation

    Returns:
        Spark dataframe with example data

    """
    try:
        findspark.init()
        spark = SparkSession.builder.getOrCreate()
        seed = seed_value
        np.random.seed(seed)
        values = []
        for i in np.arange(n_persons):
            period = np.random.randint(n_periods) + 1
            x_1 = np.random.uniform()
            x_2 = rn.choice(["A", "B", "C"])
            x_3 = np.random.uniform() + 1.0
            #Pyspark RDD does not support a column with mutliple dtypes (both string and int)
            x_4_categories = [1, 2, 3, 'a', 'b', 'c', np.nan]
            x_4 = rn.choice(x_4_categories)
            while period <= n_periods:
                values.append([i, period, x_1, x_2, x_3, x_4])
                if x_2 == "A":
                    x_1 += np.random.uniform(0, 0.1)
                else:
                    x_1 += np.random.uniform(0, 0.2)
                if x_1 > np.sqrt(x_3):
                    break
                if x_4 in x_4_categories[:-2]:
                    x_4_transition_value = x_4_categories[x_4_categories.index(x_4) + 1]
                    if np.random.uniform() >= 0.75:
                        x_4 = x_4_transition_value
                        del x_4_transition_value
                period += 1
        values = pd.DataFrame(
            values,
            columns=[
                "individual",
                "period",
                "feature_1",
                "feature_2",
                "feature_3",
                "feature_4",
            ],
        )
        values["feature_5"] = values["feature_2"]
        schema = StructType([
            StructField('individual', IntegerType(), True),
            StructField('period', IntegerType(), True),
            StructField('feature_1', FloatType(), True),
            StructField('feature_2', StringType(), True),
            StructField('feature_3', FloatType(), True),
            StructField('feature_4', StringType(), True),
            StructField('feature_5', StringType(), True)])
        return spark.createDataFrame(values, schema = schema)
    except MemoryError:
        raise MemoryError("""Dataset is too large for a pandas dataframes to store at once.
                          Consider using 'create_example_data_spark' to simulate a dataset
                          using a pyspark dataframe.""")

def import_data_file(path: str = "Input Data") -> pyspark.sql.DataFrame:
    """ Read data into a distributed spark dataframe.. 
    """
    findspark.init()
    spark = SparkSession.builder.getOrCreate()
    
    if path is not None:
        if path.endswith(".avro"):
            data = spark.read.format('avro').load(path)
        elif path.endswith(".csv"):
            data = spark.read.format('csv').load(path)
        elif path.endswith('.txt'):
            data = spark.read.format('text').load(path)
        elif path.endswith(".json"):
            data = spark.read.format('json').load(path)
        elif path.endswith(".parquet"):
            data = spark.read.load(path)
        else:
            raise TypeError(
                "Data file extension is invalid.")
    return data

class FIFEArgParser(argparse.ArgumentParser):
    """Argument parser for the FIFE command-line interface."""

    def __init__(self):
        super().__init__()
        self.add_argument(
            "--SEED",
            type=int,
            default=9999,
            help="The initializing value for all random number generators.",
        )
        self.add_argument(
            "--INDIVIDUAL_IDENTIFIER",
            type=str,
            help="The name of the feature that identifies individuals that persist over multiple time periods in the data.",
        )
        self.add_argument(
            "--TIME_IDENTIFIER",
            type=str,
            help="The name of the feature that identifies time periods in the data.",
        )
        self.add_argument(
            "--CATEGORICAL_SUFFIXES",
            type=str,
            nargs="+",
            help="Optional list of suffixes denoting that columns ending with such a suffix should be treated as categorical.",
        )
        self.add_argument(
            "--MAX_NULL_SHARE",
            type=float,
            help="The maximum share of observations that may have a null value for a feature to be kept for training.",
        )
        self.add_argument(
            "--MAX_UNIQUE_CATEGORIES",
            type=int,
            help="The maximum number of unique values for a feature of a numeric type to be considered categorical.",
        )
        self.add_argument(
            "--NUMERIC_SUFFIXES",
            type=str,
            nargs="+",
            help="Optional list of suffixes denoting that columns ending with such a suffix should be treated as numeric.",
        )
        self.add_argument(
            "--TEST_INTERVALS",
            type=int,
            help="The number of most recent periods to treat as absent from the data during training for the purpose of model evaluation.",
        )
        self.add_argument(
            "--VALIDATION_SHARE",
            type=float,
            help="The share of observations used for evaluation instead of training for hyperoptimization or early stopping.",
        )
        self.add_argument(
            "--MAX_EPOCHS",
            type=int,
            help="If HYPER_TRIALS is zero, the maximum number of passes through the training set.",
        )
        self.add_argument(
            "--MIN_SURVIVOR_IN_TRAIN",
            type=int,
            help="The minimum number of training set observations surviving a given time horizon for the model to be trained to make predictions for that time horizon.",
        )
        self.add_argument(
            "--CACHE",
            type=bool,
            help="A boolean value indicating whether or not to cache designated objects. Caching improves algorithm runtime, but is not applicable on larger datasets due to memory errors.",
        )


"""FIFE modelers based on Pyspark GBT, which trains gradient-boosted trees."""
from typing import Union

import pyspark.sql
from pyspark.ml import Pipeline
from pyspark.ml.feature import VectorAssembler, StringIndexer
from fifeforspark.base_modelers import SurvivalModeler
from fifeforspark.lgb_modelers import LGBModeler
from pyspark.ml.classification import GBTClassifier as gbt
from pyspark.sql.functions import lit
from warnings import warn

class GBTModeler(LGBModeler):

    def train_single_model(
            self,
            time_horizon: int,
            params: Union[None, dict] = None,
            subset: Union[None, pyspark.sql.DataFrame] = None
    ) -> pyspark.ml.pipeline.PipelineModel:
        """
        Train a Pyspark ML GBTClassifier model for a single lead length.

        Args:
            time_horizon: The number of periods out for which to build this model
            params: Parameters for model tuning
            subset: Boolean column for subsetting the data

        Returns:
            Single ML Pipeline model
        """
        warn("Current functionality does not support any missing values. Please remove any missing values or label"
             "them as their own category")

        if subset is None:
            subset = ~self.data[self.test_col] & ~self.data[self.predict_col]

        else:
            self.data = self.data.to_koalas()
            self.data['subset'] = subset.to_koalas()[list(subset.columns)[0]]
            self.data = self.data.to_spark()
            subset = self.data['subset']
            self.data = self.data.drop('subset')

        data = self.label_data(time_horizon)
        data = data.filter(subset)
        data = self.subset_for_training_horizon(data, time_horizon)
        
        train_data = data.filter(~data[self.validation_col])[
            self.categorical_features + self.numeric_features + [data['_label']]
            ]
        
        weight_col = self.weight_col
        if not self.weight_col:
            train_data = train_data.withColumn('weight', lit(1.0))
            weight_col = "weight"

        indexers = [StringIndexer(inputCol=column, outputCol=column + "_index").setHandleInvalid("keep")
                    for column in self.categorical_features]
        feature_columns = [column + "_index" for column in self.categorical_features] + [x for x in train_data.columns
                                                                                         if x in self.numeric_features]
        assembler = VectorAssembler(inputCols=feature_columns, outputCol='features').setHandleInvalid("keep")

        if params is None:
            gbt_model = gbt(featuresCol="features",
                            labelCol="_label",
                            weightCol=weight_col
                            )
        else:
            gbt_model = gbt(featuresCol="features",
                            labelCol="_label",
                            **params,
                            weightCol=weight_col
                            )

        pipeline = Pipeline(stages=[*indexers, assembler, gbt_model])
        model = pipeline.fit(train_data)
        return model


class GBTSurvivalModeler(GBTModeler, SurvivalModeler):
    """Use Pyspark's ML GBT Classifier to forecast probabilities of being observed in future periods."""

    pass


from time import time
from warnings import warn
from fifeforspark import processors, utils, gbt_modelers


def parse_config() -> dict:
    """
    Parse configuration parameters specified in the command line.

    Returns:
        Configuration dictionary generated from the command line input
    """

    parser = utils.FIFEArgParser()
    args = parser.parse_args()
    config = {}
    config.update({k: v for k, v in vars(args).items() if v is not None})
    return config


def main():
    """
    Executable code of FIFE when run from the command line

    Returns:
        None
    """
    warn("""The current __main__.py implementation is not meant to be a full command-line
         interface for running FIFEForSpark without writing any code, but is instead only meant 
         to test and benchmark a pre-set FIFEForSpark pipeline. Further implementation will come
         in future versions.""")
    pdp_start = time()
    config = parse_config()
    spark_df = utils.create_example_data(n_persons = 100, n_periods = 12)
    data_processor = processors.PanelDataProcessor(config, spark_df, shuffle_parts = 20)
    data_processor.build_processed_data()
    data_processor.data.show()
    pdp_end = time()
    print('PDP time:', round(pdp_end - pdp_start, 2), 'seconds.')
    
    model_train_start = time()
    modeler_class = gbt_modelers.GBTSurvivalModeler    
    modeler = modeler_class(config=config, data=data_processor.data)
    modeler.build_model()
    model_train_end = time()
    print('Model training time:', round(model_train_end - model_train_start,2), 'seconds')
    
    model_forecast_start = time()
    print(modeler.forecast())
    model_forecast_end = time()
    print('Model forecast time:', round(model_forecast_end - model_forecast_start,2), 'seconds')
    
    modeler.n_intervals = modeler.set_n_intervals()
    modeler.build_model(n_intervals=modeler.n_intervals)
    
    model_evaluate_start = time()
    print(modeler.evaluate())
    model_evaluate_end = time()
    print('Model evaluate time:', round(model_evaluate_end - model_evaluate_start,2), 'seconds')


if __name__ == '__main__':
    main()



"""FIFEforSpark modelers based on LightGBM, which trains gradient-boosted trees."""

from typing import List, Union

import pyspark.sql
from pyspark.ml import Pipeline
from pyspark.ml.feature import VectorAssembler, StringIndexer
from pyspark.sql.functions import udf, date_format, col
from pyspark.sql.types import FloatType
from fifeforspark.base_modelers import default_subset_to_all, Modeler, SurvivalModeler
import databricks.koalas as ks
from tqdm import tqdm

from warnings import warn

try:
    import mmlspark.lightgbm.LightGBMClassifier as lgb
except ImportError:
    warn("MMLSpark could not be imported. You will not be able to use LGBModeler ")


class LGBModeler(Modeler):
    """Train a gradient-boosted tree model for each lead length using MMLSpark's LightGBM.

    Attributes:
        config (dict): User-provided configuration parameters.
        data (pd.core.frame.DataFrame): User-provided panel data.
        categorical_features (list): Column names of categorical features.
        duration_col (str): Name of the column representing the number of
            future periods observed for the given individual.
        event_col (str): Name of the column indicating whether the individual
            is observed to exit the dataset.
        predict_col (str): Name of the column indicating whether the
            observation will be used for prediction after training.
        test_col (str): Name of the column indicating whether the observation
            will be used for testing model performance after training.
        validation_col (str): Name of the column indicating whether the
            observation will be used for evaluating model performance during
            training.
        period_col (str): Name of the column representing the number of
            periods since the earliest period in the data.
        max_lead_col (str): Name of the column representing the number of
            observable future periods.
        spell_col (str): Name of the column representing the number of
            previous spells of consecutive observations of the same individual.
        weight_col (str): Name of the column representing observation weights.
        reserved_cols (list): Column names of non-features.
        numeric_features (list): Column names of numeric features.
        n_intervals (int): The largest number of periods ahead to forecast.
        model (list): A trained LightGBM model (lgb.basic.Booster) for each
            lead length.
        objective (str): The LightGBM model objective appropriate for the
            outcome type, which is "binary" for binary classification.
        num_class (int): The num_class LightGBM parameter, which is 1 for
            binary classification.
    """

    def build_model(
            self,
            n_intervals: Union[None, int] = None,
            params: dict = None
    ) -> None:
        """
        Train and store a sequence of gradient-boosted tree models.

        Args:
            n_intervals: the maximum periods ahead the model will predict.
            params: Parameters for model tuning

        Returns:
            None
        """

        if n_intervals:
            self.n_intervals = n_intervals
        else:
            self.n_intervals = self.set_n_intervals()
        self.model = self.train(
            params=params
        )

    def train(
            self,
            params: Union[None, dict] = None,
            subset: Union[None, pyspark.sql.DataFrame] = None
    ) -> List[pyspark.ml.pipeline.PipelineModel]:
        """
        Train a LightGBM model for each lead length.

        Args:
            params: Parameters for model tuning
            subset: Boolean column for subsetting the data

        Returns:
            List of Pyspark ML Pipeline models
        """
        models = []
        pbar = tqdm(range(self.n_intervals))
        for time_horizon in pbar:
            pbar.set_description(f"Training models. Currently training model for time horizon {time_horizon}")
            model = self.train_single_model(
                time_horizon=time_horizon,
                params=params,
                subset=subset
            )
            models.append(model)

        return models

    def train_single_model(
            self,
            time_horizon: int,
            params: Union[None, dict] = None,
            subset: Union[None, pyspark.sql.DataFrame] = None
    ) -> pyspark.ml.pipeline.PipelineModel:
        """
        Train a LightGBM model for a single lead length.

        Args:
            time_horizon: The number of periods out for which to build this model
            params: Parameters for model tuning
            subset: Boolean column for subsetting the data

        Returns:
            Single ML Pipeline model
        """
        if params is None:
            params = {
                time_horizon: {
                    "objective": self.objective,
                    "numIterations": self.config.get("MAX_EPOCHS", 256),
                }
            }
        elif params.get(time_horizon, None) is None:
            params[time_horizon] = {
                    "objective": self.objective,
                    "numIterations": self.config.get("MAX_EPOCHS", 256),
                }
        if subset is None:
            subset = ~self.data[self.test_col] & ~self.data[self.predict_col]

        else:
            self.data = self.data.to_koalas()
            self.data['subset'] = subset.to_koalas()[list(subset.columns)[0]]
            self.data = self.data.to_spark()
            subset = self.data['subset']
            self.data = self.data.drop('subset')

        data = self.label_data(time_horizon)
        data = data.filter(subset)
        data = self.subset_for_training_horizon(data, time_horizon)

        train_data = data.filter(~data[self.validation_col])[
            self.categorical_features + self.numeric_features + [data['_label']]
            ] 
        indexers = [StringIndexer(inputCol=column, outputCol=column + "_index").setHandleInvalid("keep")
                    for column in self.categorical_features]
        feature_columns = [column + "_index" for column in self.categorical_features] + self.numeric_features
        assembler = VectorAssembler(inputCols=feature_columns, outputCol='features').setHandleInvalid("keep")
        lgb_model = lgb(featuresCol="features",
                        labelCol="_label",
                        **params[time_horizon],
                        weightCol=data.filter(~data[self.validation_col])[self.weight_col]
                        if self.weight_col
                        else None
                        )
        pipeline = Pipeline(stages=[*indexers, assembler, lgb_model])
        model = pipeline.fit(train_data)
        return model

    def predict(
            self, subset: Union[None, pyspark.sql.DataFrame] = None, cumulative: bool = True
    ) -> pyspark.sql.DataFrame:
        """Use trained LightGBM models to predict the outcome for each observation and time horizon.

        Args:
            subset: A Boolean Spark Column that is True for observations for which
                predictions will be produced. If None, default to all
                observations.
            cumulative: If True, produce cumulative survival probabilities.
                If False, produce marginal survival probabilities (i.e., one
                minus the hazard rate).

        Returns:
            A dataframe of predictions by observation and lead
            length.
        """
        subset = default_subset_to_all(subset, self.data)

        self.data = self.data.to_koalas()
        self.data['subset'] = subset.to_koalas()[list(subset.columns)[0]]
        self.data = self.data.to_spark()

        predict_data = self.data.filter(self.data['subset'])[self.categorical_features + self.numeric_features]

        secondelement = udf(lambda v: float(v[1]), FloatType())
        first_model = self.model[0]
        predictions = first_model.transform(predict_data).selectExpr('probability as probability_1')
        predictions = predictions.withColumn('probability_1', secondelement(predictions['probability_1']))
        predictions = predictions.to_koalas()
        for i, lead_specific_model in enumerate(self.model):
            if i != 0:
                pred_year = lead_specific_model.transform(predict_data).selectExpr(f'probability as probability_{i+1}')
                predictions[f'probability_{i+1}'] = pred_year.select(secondelement(pred_year[f'probability_{i+1}'])).to_koalas()
                if cumulative:
                    predictions[f'probability_{i + 1}'] = predictions[f'probability_{i + 1}'] * predictions[f'probability_{i}']
        return predictions.to_spark()

    def transform_features(self) -> pyspark.sql.DataFrame:
        """
        Transform datetime features to suit model training.

        Returns:
            Spark DataFrame with transformed features
        """
        data = self.data
        date_cols = [x for x, y in data.dtypes if y in ['date', 'timestamp']]
        for date_col in date_cols:
            data = data.withColumn(date_col,
                                   10000*date_format(data[date_col], "y") +
                                   100*date_format(data[date_col], "M") +
                                   date_format(data[date_col], "d"))
        return data

    def save_model(self, path: str = "") -> None:
        """
        Save the horizon-specific LightGBM models that comprise the model to disk.
        Functionality currently in progress.

        Args:
            file_name: The desired name of the model on disk
            path: The path for where to save the model

        Returns:
            None
        """
        for i, lead_specific_model in enumerate(self.model):
            lead_specific_model.write().overwrite().save(f'{path}_model{i}')
        raise NotImplementedError("Model Saving functionality not yet completed.")


class LGBSurvivalModeler(LGBModeler, SurvivalModeler):
    """Use LightGBM to forecast probabilities of being observed in future periods."""

    pass



from abc import ABC, abstractmethod
from typing import Union, Any
from collections import OrderedDict

import pandas as pd
import numpy as np
import databricks.koalas as ks
import findspark
import pyspark.sql
from pyspark.sql import SparkSession
from pyspark.sql.types import DoubleType
from pyspark.sql.functions import lit, when, monotonically_increasing_id, mean, col
from pyspark.ml.evaluation import BinaryClassificationEvaluator
from tqdm import tqdm
from sklearn.metrics import confusion_matrix, r2_score, roc_auc_score


def default_subset_to_all(
    subset: Union[None, pyspark.sql.DataFrame], data: pyspark.sql.DataFrame
) -> pyspark.sql.DataFrame:
    """
    Map an unspecified subset to an entirely True boolean mask.

    Args:
        subset: the subset passed into the data (boolean column) if it exists
        data: the original dataset

    Returns:
        Original subset argument if not none, otherwise a new boolean mask that is always True.
    """
    if subset is None:
        return data.withColumn('True_mask', lit(True)).select('True_mask')
    else:
        assert len(subset.columns) == 1, "Provided subset is not a valid one-column subset"
    return subset


def compute_metrics_for_binary_outcome(
    actuals: pyspark.sql.DataFrame, predictions: pyspark.sql.DataFrame, total: int,
    threshold_positive: Union[None, str, float] = 0.5, share_positive: Union[None, str, float] = None,
    cache = False
) -> OrderedDict:
    """
    Function to compute performance metrics for binary classification models

    Args:
        actuals: The actual values
        predictions: The predicted values
        threshold_positive: The threshold for determining whether a value is predicted True or False
        share_positive: The share of values that you want to classify as positive

    Returns:
        Ordered dictionary with evaluation metrics
    """
    if cache is True:
        predictions.cache()
        actuals.cache()
    actuals = actuals.select(actuals['actuals'].cast(DoubleType()))
    num_true = actuals.agg({'actuals': 'sum'}).first()[0]
    metrics = OrderedDict()
    if num_true is None:
        num_true = 0
    if (num_true > 0) & (num_true < total):
        predictions = predictions.withColumn(
            "row_id", monotonically_increasing_id())
        actuals = actuals.withColumn("row_id", monotonically_increasing_id())
        preds_and_labs = predictions.join(actuals, predictions.row_id == actuals.row_id).select(
            predictions.predictions, actuals.actuals)
        if cache is True:
            preds_and_labs.cache()
        preds_and_labs = preds_and_labs.withColumn('rawPrediction', preds_and_labs.predictions.cast(DoubleType()))
        evaluator = BinaryClassificationEvaluator(labelCol='actuals')
        metrics['AUROC'] = evaluator.evaluate(preds_and_labs, {evaluator.metricName: "areaUnderROC"})
        preds_and_labs = preds_and_labs.drop('rawPrediction')
    else:
        metrics["AUROC"] = np.nan

    # TODO: Add weighted average functionality
    mean_predict = predictions.select(mean(col('predictions'))).first()[0]
    metrics["Actual Share"] = actuals.select(mean(col('actuals'))).first()[0]
    metrics["Predicted Share"] = mean_predict
    if actuals.first() is not None:
        if share_positive == 'predicted':
            share_positive = predictions.agg(
                {'predictions': 'mean'}).first()[0]
        if share_positive is not None:
            threshold_positive = predictions.approxQuantile(
                'predictions', [1-share_positive], relativeError=.1)[0]
        elif threshold_positive == 'predicted':
            threshold_positive = mean_predict
        preds_and_labs = preds_and_labs.withColumn('predictions', when(
            preds_and_labs.predictions >= threshold_positive, 1).otherwise(0))
        outcomes = {'True Positives': [1,1], 'False Negatives': [0,1], 'False Positives': [1,0], 'True Negatives': [0,0]}
        for key in outcomes.keys():
            metrics[key] = preds_and_labs.select(((preds_and_labs.predictions == outcomes[key][0]) & (
                preds_and_labs.actuals == outcomes[key][1])).cast('int').alias('Outcome')).agg({'Outcome': 'sum'}).first()[0]
    return metrics


class Modeler(ABC):
    """Set template for modelers that use panel data to produce forecasts.
    Attributes:
        config (dict): User-provided configuration parameters.
        data (pd.core.frame.DataFrame): User-provided panel data.
        categorical_features (list): Column names of categorical features.
        duration_col (str): Name of the column representing the number of
            future periods observed for the given individual.
        event_col (str): Name of the column indicating whether the individual
            is observed to exit the dataset.
        predict_col (str): Name of the column indicating whether the
            observation will be used for prediction after training.
        test_col (str): Name of the column indicating whether the observation
            will be used for testing model performance after training.
        validation_col (str): Name of the column indicating whether the
            observation will be used for evaluating model performance during
            training.
        period_col (str): Name of the column representing the number of
            periods since the earliest period in the data.
        max_lead_col (str): Name of the column representing the number of
            observable future periods.
        spell_col (str): Name of the column representing the number of
            previous spells of consecutive observations of the same individual.
        weight_col (str): Name of the column representing observation weights.
        reserved_cols (list): Column names of non-features.
        numeric_features (list): Column names of numeric features.
        n_intervals (int): The largest number of periods ahead to forecast
        allow_gaps (bool): Whether or not observations should be included for
            training and evaluation if there is a period without an observation
            between the period of observations and the last period of the
            given time horizon.
    """

    def __init__(
        self,
        config: Union[None, dict] = {},
        data: Union[None, pyspark.sql.DataFrame] = None,
        duration_col: str = "_duration",
        event_col: str = "_event_observed",
        predict_col: str = "_predict_obs",
        test_col: str = "_test",
        validation_col: str = "_validation",
        period_col: str = "_period",
        max_lead_col: str = "_maximum_lead",
        spell_col: str = "_spell",
        weight_col: Union[None, str] = None,
        allow_gaps: bool = False,
    ) -> None:
        """Characterize data for modelling.
        Identify each column as categorical, reserved, and numeric.
        Compute the maximum lead length for modelling.
        Args:
            config: User-provided configuration parameters.
            data: User-provided panel data.
            duration_col: Name of the column representing the number of future
                periods observed for the given individual.
            event_col: Name of the column indicating whether the individual is
                observed to exit the dataset.
            predict_col: Name of the column indicating whether the observation
                will be used for prediction after training.
            test_col: Name of the column indicating whether the observation
                will be used for testing model performance after training.
            validation_col: Name of the column indicating whether the
                observation will be used for evaluating model performance
                during training.
            period_col (str): Name of the column representing the number of
                periods since the earliest period in the data.
            max_lead_col (str): Name of the column representing the number of
                observable future periods.
            spell_col (str): Name of the column representing the number of
                previous spells of consecutive observations of the same individual.
            weight_col (str): Name of the column representing observation weights.
            allow_gaps (bool): Whether or not observations should be included for
                training and evaluation if there is a period without an observation
                between the period of observations and the last period of the
                given time horizon.
        """
        findspark.init()
        
        ks.set_option('compute.ops_on_diff_frames', True)
        
        if (config.get("TIME_IDENTIFIER", "") == "") and data is not None:
            config["TIME_IDENTIFIER"] = data.columns[1]

        if (config.get("INDIVIDUAL_IDENTIFIER", "") == "") and data is not None:
            config["INDIVIDUAL_IDENTIFIER"] = data.columns[0]

        self.config = config
        self.data = data
        self.duration_col = duration_col
        self.event_col = event_col
        self.predict_col = predict_col
        self.test_col = test_col
        self.validation_col = validation_col
        self.period_col = period_col
        self.max_lead_col = max_lead_col
        self.spell_col = spell_col
        self.weight_col = weight_col
        self.allow_gaps = allow_gaps
        self.reserved_cols = [
            self.duration_col,
            self.event_col,
            self.predict_col,
            self.test_col,
            self.validation_col,
            self.period_col,
            self.max_lead_col,
            self.spell_col,
        ]
        if self.config:
            self.reserved_cols.append(self.config["INDIVIDUAL_IDENTIFIER"])
        if self.weight_col:
            self.reserved_cols.append(self.weight_col)
        if self.data is not None:
            self.categorical_features = [
                col[0] for col in self.data.dtypes if col[1] == 'string']
            self.numeric_features = [feature for feature in self.data.columns
                                     if feature not in (self.categorical_features + self.reserved_cols)]
            self.data = self.transform_features()

    @abstractmethod
    def train(self) -> Any:
        """
        Train and return a model.

        Returns:
            Any
        """

    @abstractmethod
    def predict(
        self, subset: Union[None, pyspark.sql.DataFrame] = None, cumulative: bool = True
    ) -> pyspark.sql.DataFrame:
        """
        Use trained model to produce observation survival probabilities.

        Args:
            subset: The subset of values to predict on
            cumulative: Flag for whether output probabilities should be cumulative

        Returns:
            Spark dataframe with prediction results
        """

    @abstractmethod
    def evaluate(
        self,
        subset: Union[None, pyspark.sql.DataFrame] = None,
        threshold_positive: Union[None, str, float] = 0.5,
        share_positive: Union[None, str, float] = None,
    ) -> pd.core.frame.DataFrame:
        """
        Tabulate model performance metrics.

        Args:
            subset: The subset of values to evaluate
            threshold_positive: The threshold value for determining whether a value is positive or negative
            share_positive: The share you want to classify as positive

        Returns:
            Pandas Dataframe with evaluation results
        """

    @abstractmethod
    def forecast(self) -> pyspark.sql.DataFrame:
        """
        Tabulate survival probabilities for most recent observations.

        Returns:
            Spark DataFrame with forecast results
        """

    @abstractmethod
    def subset_for_training_horizon(
        self, data: pyspark.sql.DataFrame, time_horizon: int
    ) -> pyspark.sql.DataFrame:
        """
        Return only observations where the outcome is observed.

        Args:
            data: Dataset to train on
            time_horizon: the number of periods for which you're forecasting (e.g. 2 periods out)

        Returns:
            Spark DataFrame with original dataset subsetted
        """

    @abstractmethod
    def label_data(self, time_horizon: int) -> pyspark.sql.DataFrame:
        """
        Return data with an outcome label for each observation.

        Args:
            time_horizon: the number of periods for which you're forecasting (e.g. 2 periods out)

        Returns:
            Data with outcome label added based on time_horizon
        """

    @abstractmethod
    def save_model(self, path: str = "") -> None:
        """
        Save model file(s) to disk.

        Args:
            path: Path to save model

        Returns:
            None
        """

    @abstractmethod
    def transform_features(self) -> pyspark.sql.DataFrame:
        """
        Transform datetime features to suit model training.

        Returns:
            Spark DataFrame with transformed features
        """

    @abstractmethod
    def build_model(self, n_intervals: Union[None, int] = None) -> None:
        """
        Configure, train, and store a model.

        Args:
            n_intervals: the maximum periods ahead the model will predict.

        Returns:
            None
        """

    def set_n_intervals(self) -> int:
        """
        Determine the maximum periods ahead the model will predict.

        Returns:
            n_intervals (the maximum periods ahead the model will predict)
        """
        train_durations = self.data.select(
            (when(
                self.data[self.duration_col] <= self.data[self.max_lead_col], self.data[self.duration_col]
            ).otherwise(self.data[self.max_lead_col])).alias('max_durations')
        )
        subset = ~self.data[self.validation_col] & ~self.data[self.test_col] & ~self.data[self.predict_col]
        train_obs_by_lead_length = train_durations.filter(subset)
        train_obs_by_lead_length = train_obs_by_lead_length.groupBy(
            'max_durations').count()
        min_survivors_subset = train_obs_by_lead_length.filter(
            (train_obs_by_lead_length['count'] > self.config.get(
                "MIN_SURVIVOR_IN_TRAIN", 64)).alias('max_durations'))
        assert min_survivors_subset.first() is not None, "No lead length has more than 64 survivors."
        n_intervals = min_survivors_subset.agg(
            {'max_durations': 'max'}
        ).first()[0]
        return n_intervals


class SurvivalModeler(Modeler):
    """Forecast probabilities of being observed in future periods.

    Attributes:
        config (dict): User-provided configuration parameters.
        data (pd.core.frame.DataFrame): User-provided panel data.
        categorical_features (list): Column names of categorical features.
        duration_col (str): Name of the column representing the number of
            future periods observed for the given individual.
        event_col (str): Name of the column indicating whether the individual
            is observed to exit the dataset.
        predict_col (str): Name of the column indicating whether the
            observation will be used for prediction after training.
        test_col (str): Name of the column indicating whether the observation
            will be used for testing model performance after training.
        validation_col (str): Name of the column indicating whether the
            observation will be used for evaluating model performance during
            training.
        period_col (str): Name of the column representing the number of
            periods since the earliest period in the data.
        max_lead_col (str): Name of the column representing the number of
            observable future periods.
        spell_col (str): Name of the column representing the number of
            previous spells of consecutive observations of the same individual.
        weight_col (str): Name of the column representing observation weights.
        reserved_cols (list): Column names of non-features.
        numeric_features (list): Column names of numeric features.
        n_intervals (int): The largest number of periods ahead to forecast.
        allow_gaps (bool): Whether or not observations should be included for
            training and evaluation if there is a period without an observation
            between the period of observations and the last period of the
            given time horizon.
    """

    def __init__(self, **kwargs):
        """
        Initialize the SurvivalModeler.

        Args:
            **kwargs: Arguments to Modeler.__init__().
        """
        super().__init__(**kwargs)
        self.objective = "binary"
        self.num_class = 1

    def evaluate(
        self,
        subset: Union[None, pyspark.sql.DataFrame] = None,
        threshold_positive: Union[None, str, float] = 0.5,
        share_positive: Union[None, str, float] = None,
    ) -> pd.core.frame.DataFrame:
        """
        Tabulate model performance metrics.

        Args:
            subset: A Boolean Series that is True for observations over which
                the metrics will be computed. If None, default to all test
                observations in the earliest period of the test set.
            threshold_positive: None, "predicted", or a value in [0, 1]
                representing the minimum predicted probability considered to be a
                positive prediction; Specify "predicted" to use the predicted share
                positive in each time horizon. Overridden by share_positive.
            share_positive: None, "predicted", or a value in [0, 1] representing
                the share of observations with the highest predicted probabilities
                considered to have positive predictions. Specify "predicted" to use
                the predicted share positive in each time horizon. Probability ties
                may cause share of positives in output to exceed given value.
                Overrides threshold_positive.

        Returns:
            A DataFrame containing, for the binary outcomes of survival to each
            lead length, area under the receiver operating characteristic
            curve (AUROC), predicted and actual shares of observations with an
            outcome of True, and all elements of the confusion matrix. Also
            includes concordance index over the restricted mean survival time.
        """
        try:
            return self._evaluate_single_node(subset=subset,
                                             threshold_positive=threshold_positive,
                                             share_positive=share_positive)
        except MemoryError:
            return self._evaluate_multi_node(subset=subset,
                                            threshold_positive=threshold_positive,
                                            share_positive=share_positive)

    def _evaluate_single_node(self,
        subset: Union[None, pyspark.sql.DataFrame] = None,
        threshold_positive: Union[None, str, float] = 0.5,
        share_positive: Union[None, str, float] = None,
    ) -> pd.core.frame.DataFrame:
        """
        Evaluate method for single node. Taken directly from FIFE.

        Args:
            subset: A Boolean Series that is True for observations over which
                the metrics will be computed. If None, default to all test
                observations in the earliest period of the test set.
            threshold_positive: None, "predicted", or a value in [0, 1]
                representing the minimum predicted probability considered to be a
                positive prediction; Specify "predicted" to use the predicted share
                positive in each time horizon. Overridden by share_positive.
            share_positive: None, "predicted", or a value in [0, 1] representing
                the share of observations with the highest predicted probabilities
                considered to have positive predictions. Specify "predicted" to use
                the predicted share positive in each time horizon. Probability ties
                may cause share of positives in output to exceed given value.
                Overrides threshold_positive.

        Returns:
            A DataFrame containing, for the binary outcomes of survival to each
            lead length, area under the receiver operating characteristic
            curve (AUROC), predicted and actual shares of observations with an
            outcome of True, and all elements of the confusion matrix.
        """
        def compute_metrics_for_binary_outcome_single_node(
                actuals: Union[pd.Series, pd.DataFrame],
                predictions: np.ndarray,
                threshold_positive: Union[None, str, float] = 0.5,
                share_positive: Union[None, str, float] = None,
                weights: Union[None, np.ndarray] = None,
        ) -> OrderedDict:
            """Evaluate predicted probabilities against actual binary outcome values.

            Args:
                actuals: A Series representing actual Boolean outcome values.
                predictions: A Series of predicted probabilities of the respective
                    outcome values.
                threshold_positive: None, "predicted", or a value in [0, 1]
                    representing the minimum predicted probability considered to be a
                    positive prediction; Specify "predicted" to use the predicted share
                    positive in each time horizon. Overridden by share_positive.
                share_positive: None, "predicted", or a value in [0, 1] representing
                    the share of observations with the highest predicted probabilities
                    considered to have positive predictions. Specify "predicted" to use
                    the predicted share positive in each time horizon. Probability ties
                    may cause share of positives in output to exceed given value.
                    Overrides threshold_positive.
                weights: A 1-D array of weights with the same length as actuals and predictions.
                Each prediction contributes to metrics in proportion to its weight. If None, each prediction has the same weight.

            Returns:
                An ordered dictionary containing key-value pairs for area under the
                receiver operating characteristic curve (AUROC), predicted and
                actual shares of observations with an outcome of True, and all
                elements of the confusion matrix.
            """
            metrics = OrderedDict()
            if actuals.any() and not actuals.all():
                metrics["AUROC"] = roc_auc_score(actuals, predictions, sample_weight=weights)
            else:
                metrics["AUROC"] = np.nan
            metrics["Predicted Share"] = np.average(predictions, weights=weights)
            metrics["Actual Share"] = np.average(actuals, weights=weights)
            if actuals.empty:
                (
                    metrics["True Positives"],
                    metrics["False Negatives"],
                    metrics["False Positives"],
                    metrics["True Negatives"],
                ) = [0, 0, 0, 0]
            else:
                if share_positive == "predicted":
                    share_positive = metrics["Predicted Share"]
                if share_positive is not None:
                    threshold_positive = np.quantile(predictions, 1 - share_positive)
                elif threshold_positive == "predicted":
                    threshold_positive = metrics["Predicted Share"]
                (
                    metrics["True Positives"],
                    metrics["False Negatives"],
                    metrics["False Positives"],
                    metrics["True Negatives"],
                ) = (
                    confusion_matrix(
                        actuals,
                        predictions >= threshold_positive,
                        labels=[True, False],
                        sample_weight=weights,
                    ).ravel()
                     .tolist()
                )
            return metrics

        if subset is None:
            filtered = self.data.filter(self.data[self.test_col])
            min_val = filtered.select(self.data[self.period_col]).agg(
                {self.period_col: 'min'}).first()[0]
            self.data = self.data.withColumn('subset', self.data[self.test_col] &
                                                   (self.data[self.period_col] == min_val))
            subset = self.data.select('subset')

        predictions = self.predict(subset=subset, cumulative=(not self.allow_gaps))

        metrics = []
        lead_lengths = np.arange(self.n_intervals) + 1
        for lead_length in tqdm(lead_lengths, desc="Evaluating Model by Lead Length"):
            lead_length = int(lead_length)
            labeled_data = self.label_data(int(lead_length - 1))
            if 'subset' not in labeled_data.columns:
                labeled_data = labeled_data.to_koalas()
                labeled_data['subset'] = subset.to_koalas()[list(subset.columns)[0]]
                labeled_data = labeled_data.to_spark()

            actuals = labeled_data.filter(labeled_data['subset'])
            actuals = actuals.filter(actuals[self.max_lead_col] >= lead_length)
            weights = actuals.select(self.weight_col).collect() if self.weight_col else None
            actuals = actuals.select("_label").toPandas()['_label']
            metrics.append(
                compute_metrics_for_binary_outcome_single_node(
                    actuals,
                    np.array(predictions.select(predictions.columns[lead_length-1]).collect()),
                    threshold_positive=threshold_positive,
                    share_positive=share_positive,
                    weights=weights,
                )
            )
        metrics = pd.DataFrame(metrics, index=lead_lengths)
        metrics.index.name = "Lead Length"
        metrics = metrics.dropna()
        return metrics

    def _evaluate_multi_node(self,
        subset: Union[None, pyspark.sql.DataFrame] = None,
        threshold_positive: Union[None, str, float] = 0.5,
        share_positive: Union[None, str, float] = None,
    ) -> pd.core.frame.DataFrame:

        """
        Args:
            subset: A Boolean Series that is True for observations over which
                the metrics will be computed. If None, default to all test
                observations in the earliest period of the test set.
            threshold_positive: None, "predicted", or a value in [0, 1]
                representing the minimum predicted probability considered to be a
                positive prediction; Specify "predicted" to use the predicted share
                positive in each time horizon. Overridden by share_positive.
            share_positive: None, "predicted", or a value in [0, 1] representing
                the share of observations with the highest predicted probabilities
                considered to have positive predictions. Specify "predicted" to use
                the predicted share positive in each time horizon. Probability ties
                may cause share of positives in output to exceed given value.
                Overrides threshold_positive.

        Returns:
            A DataFrame containing, for the binary outcomes of survival to each
            lead length, area under the receiver operating characteristic
            curve (AUROC), predicted and actual shares of observations with an
            outcome of True, and all elements of the confusion matrix. Also
            includes concordance index over the restricted mean survival time.
        """

        predictions = self.predict(subset=subset, cumulative=(not self.allow_gaps))
        lead_lengths = np.arange(self.n_intervals) + 1
        metrics = []
        total = -1
        for lead_length in tqdm(lead_lengths, desc="Evaluating Model by Lead Length"):
            actuals = self.label_data(int(lead_length - 1))
            if subset is None:
                min_val = actuals.select(actuals['_period']).agg({'_period': 'min'}).first()[0]
                actuals = actuals.withColumn('subset', actuals[self.test_col] &
                                             (actuals[self.period_col] == min_val))
            else:
                actuals = actuals.to_koalas()
                actuals['subset'] = subset.to_koalas()[list(subset.columns)[0]]
                actuals = actuals.to_spark()

            actuals = actuals.filter(actuals.subset)
            actuals = actuals.filter(actuals[self.max_lead_col] >= int(lead_length))
            actuals = actuals.select(actuals["_label"].alias('actuals'))
            if lead_length == 1:
                total = actuals.count()
            metrics.append(
                compute_metrics_for_binary_outcome(
                    actuals,
                    predictions.select(
                        predictions[int(lead_length - 1)].alias('predictions')),
                    total=total,
                    threshold_positive=threshold_positive,
                    share_positive=share_positive,
                    cache=self.config.get('CACHE', False)
                )
            )
        metrics = pd.DataFrame(metrics, index=lead_lengths)
        metrics.index.name = "Lead Length"
        metrics = metrics.dropna()
        return metrics

    def forecast(self) -> ks.DataFrame:
        """
        Tabulate survival probabilities for most recent observations.

        Returns:
            Spark DataFrame with forecast results
        """
        columns = [
            str(i + 1) + "-period Survival Probability" for i in range(self.n_intervals)]
        forecasts = self.predict(
            subset=self.data.select(self.predict_col), cumulative=(not self.allow_gaps))
        forecasts = forecasts.to_koalas()
        forecasts.columns = columns
        forecasts['Index'] = self.data.filter(self.predict_col).select(self.data[self.config["INDIVIDUAL_IDENTIFIER"]]).to_koalas()
        forecasts = forecasts.set_index('Index')
        # TODO: Add custom index values
        return forecasts

    def subset_for_training_horizon(self, data: pyspark.sql.DataFrame, time_horizon: int) -> pyspark.sql.DataFrame:
        """
        Return only observations where the outcome is observed.

        Args:
            data: Dataset to train on
            time_horizon: the number of periods for which you're forecasting (e.g. 2 periods out)

        Returns:
            Spark DataFrame with original dataset subsetted
        """
        if self.allow_gaps:
            return data.filter(data[self.max_lead_col] > time_horizon)
        return data.filter((data[self.duration_col] + data[self.event_col].cast('int')) > time_horizon)

    def label_data(self, time_horizon: int) -> pyspark.sql.DataFrame:
        """
        Return data with an outcome label for each observation.

        Args:
            time_horizon: the number of periods for which you're forecasting (e.g. 2 periods out)

        Returns:
            Data with outcome label added based on time_horizon
        """
        spark_df = self.data
        spark_df = spark_df.withColumn(self.duration_col, when(
            spark_df[self.duration_col] <= spark_df[self.max_lead_col], spark_df[self.duration_col]).otherwise(spark_df[self.max_lead_col]))
        if self.allow_gaps:
            ids = spark_df[self.config["INDIVIDUAL_IDENTIFIER"],
                           self.config["TIME_IDENTIFIER"]]
            ids = ids.withColumn(
                self.config["INDIVIDUAL_IDENTIFIER"] + '_new', ids[self.config["INDIVIDUAL_IDENTIFIER"]])
            ids = ids.withColumn(
                self.config["TIME_IDENTIFIER"] + '_new', ids[self.config["TIME_IDENTIFIER"]] - time_horizon - 1)
            ids = ids.withColumn('_label', lit(True))
            ids = ids.drop(
                self.config["INDIVIDUAL_IDENTIFIER"], self.config["TIME_IDENTIFIER"])

            spark_df = spark_df.join(ids, (spark_df[self.config["INDIVIDUAL_IDENTIFIER"]] == ids[self.config["INDIVIDUAL_IDENTIFIER"] + '_new']) & (
                spark_df[self.config["TIME_IDENTIFIER"]] == ids[self.config["TIME_IDENTIFIER"] + '_new']), "left")
            spark_df = spark_df.drop(
                self.config["INDIVIDUAL_IDENTIFIER"] + '_new', self.config["TIME_IDENTIFIER"] + '_new')
            spark_df = spark_df.fillna(False, subset=['_label'])
        else:
            spark_df = spark_df.withColumn(
                '_label', (spark_df[self.duration_col] > lit(time_horizon)).cast('int'))
        return spark_df


class StateModeler(Modeler):
    def __init__(self):
        raise NotImplementedError("StateModeler not yet implemented.")


class ExitModeler(StateModeler):
    def __init__(self):
        raise NotImplementedError("ExitModeler not yet implemented.")



import findspark
import pyspark
import pyspark.sql
import pyspark.sql.functions as F
import databricks.koalas as ks
from pyspark.sql import SparkSession, Window
from pyspark.sql.functions import isnan, lit, col, lag
from pyspark.sql.types import DateType, TimestampType, StringType, IntegerType, LongType, ShortType, ByteType, FloatType, DoubleType, DecimalType
from typing import Union
from warnings import warn


class DataProcessor:
    """Prepare data by identifying features as degenerate or categorical."""

    def __init__(self, config={}, data=None) -> None:
        """Initialize the DataProcessor.

        Args:
            config: A dictionary of configuration parameters.
            data: A DataFrame to be processed.

        """
        findspark.init()
        self.spark = SparkSession.builder.getOrCreate()
        if (config.get("INDIVIDUAL_IDENTIFIER", "") == "") and data is not None:
            config["INDIVIDUAL_IDENTIFIER"] = data.columns[0]
            print(
                "Individual identifier column name not given; assumed to be "
                f'leftmost column ({config["INDIVIDUAL_IDENTIFIER"]})'
            )
        self.config = config
        self.data = data

    def check_column_consistency(self, colname: str) -> None:
        """
        Assert column exists, has no missing values, and is not constant.

        Args:
            colname: The name of the column to check

        Returns:
            None
        """
        if colname == self.config['TIME_IDENTIFIER']:
            ks_df = ks.DataFrame(self.data)
            ks_df['_period']= ks_df[self.config['TIME_IDENTIFIER']].factorize(sort=True)[0]
            self.data = ks_df.to_spark()
            colname = '_period'
        assert colname in self.data.columns, f"{colname} not in data"
        assert self.data.select(
            isnan(colname).cast('int').alias(colname)
        ).agg({colname: 'sum'}).first()[0] == 0, f"{colname} has missing values"
        assert self.data.select(
            colname).distinct().count() >= 2, f"{colname} does not have multiple unique values"

    def is_degenerate(self, colname: str) -> bool:
        """
        Determine if a feature is constant or has too many missing values

        Args:
            col: The column/feature to check

        Returns:
            Boolean value for whether the column is degenerate
        """
        if colname == self.config['TIME_IDENTIFIER']:
            ks_df = ks.DataFrame(self.data)
            ks_df['_period']= ks_df[self.config['TIME_IDENTIFIER']].factorize(sort=True)[0]
            self.data = ks_df.to_spark()
            colname = '_period'
        if self.data.select(
                isnan(colname).cast('integer').alias(colname)
        ).agg({colname: 'mean'}).first()[0] >= self.config.get('MAX_NULL_SHARE', 0.999):
            return True
        if self.data.select(colname).distinct().count() < 2:
            return True
        return False

    def is_categorical(self, col: str) -> bool:
        """
        Determine if the given feature should be processed as categorical, as opposed to numeric.

        Args:
            col: The column to check

        Returns:
            Boolean value for whether the column is categorical
        """
        try:
            if col.endswith(self.config.get('CATEGORICAL_SUFFIXES', ())):
                if col.endswith(self.config.get('NUMERIC_SUFFIXES', ())):
                    print(
                        f"{col} matches both categorical and numerical suffixes; it will be identified as categorical"
                    )
                return True
        except TypeError:
            raise TypeError("'NUMERIC_SUFFIXES' and 'CATEGORICAL_SUFFIXES' must be either strings or tuples of strings.")
        if isinstance(self.data.schema[col].dataType, (DateType, TimestampType)):
            return False
        if not isinstance(self.data.schema[col].dataType, (IntegerType, LongType, ShortType,
                                                           ByteType, FloatType, DoubleType, DecimalType)):
            if col.endswith(self.config.get('NUMERIC_SUFFIXES', ())):
                print(
                    f"{col} matches numeric suffix but is non-numeric; identified as categorical"
                )
            return True
        if (col.endswith(self.config.get('NUMERIC_SUFFIXES', ()))) or (
                self.data.select(col).distinct().count() > self.config.get('MAX_UNIQUE_CATEGORIES', 1024)):
            return False
        return True


class PanelDataProcessor(DataProcessor):
    """Ready panel data for modelling.

    Attributes:
        config (dict): User-provided configuration parameters.
        data (pd.core.frame.DataFrame): Processed panel data.
        raw_subset (pd.core.frame.DataFrame): An unprocessed sample from the
            final period of data. Useful for displaying meaningful values in
            SHAP plots.
        categorical_maps (dict): Contains for each categorical feature a map
            from each unique value to a whole number.
        numeric_ranges (pd.core.frame.DataFrame): Contains for each numeric
            feature the maximum and minimum value in the training set.
    """

    def __init__(self, config: Union[None, dict] = {},
                 data: Union[None, pyspark.sql.DataFrame] = None, shuffle_parts = 200) -> None:
        """Initialize the PanelDataProcessor.

        Args:
            config: A dictionary of configuration parameters.
            data: A DataFrame to be processed.
        """

        if (config.get("TIME_IDENTIFIER", "") == "") and data is not None:
            config["TIME_IDENTIFIER"] = data.columns[1]
            print(
                "Time identifier column name not given; assumed to be "
                f'second-leftmost column ({config["TIME_IDENTIFIER"]})'
            )
        super().__init__(config, data)
        self.spark.conf.set("spark.sql.shuffle.partitions", shuffle_parts)

    def check_panel_consistency(self) -> None:
        """
        Ensure observations have unique individual-period combinations.

        Returns:
            None
        """
        self.check_column_consistency(self.config['INDIVIDUAL_IDENTIFIER'])
        self.check_column_consistency(self.config['TIME_IDENTIFIER'])
        subset = self.data.select(
            [self.config['INDIVIDUAL_IDENTIFIER'], self.config['TIME_IDENTIFIER']])
        assert subset.count() == subset.dropDuplicates().count(
        ), "One or more individuals have multiple observations for a single time value."

    def process_single_column(self, colname) -> pyspark.sql.DataFrame:
        """
        Apply data cleaning functions to a singular data column.

        Args:
            colname: The column to process

        Returns:
            Dataframe: dataframe with the processed column
        """
        if colname == self.config['INDIVIDUAL_IDENTIFIER']:
            return self.data
        if self.is_degenerate(colname=colname):
            self.data = self.data.drop(colname)
        elif self.is_categorical(colname):
            self.data = self.data.withColumn(
                colname, self.data[colname].cast(StringType()))
            self.data = self.data.fillna('NaN', subset=[colname])
        return self.data

    def process_all_columns(self) -> pyspark.sql.DataFrame:
        """
        Apply data cleaning functions to all data columns.

        Returns:
            Spark DataFrame with processed columns
        """
        for col_name in self.data.columns:
            self.data = self.process_single_column(colname=col_name)
        return self.data

    def flag_validation_individuals(self) -> pyspark.sql.DataFrame:
        """
        Flag observations from a random share of individuals.

        Returns:
            Spark DataFrame with flagged validation individuals
        """
        unique_ids = self.data.select(
            self.config['INDIVIDUAL_IDENTIFIER']).distinct()
        total = unique_ids.count()
        val_size = int(self.config.get('VALIDATION_SHARE', 0.25) * total)
        unique_ids = unique_ids.sample(True, val_size/total, seed=9999)
        unique_list = list(
            unique_ids.select(self.config['INDIVIDUAL_IDENTIFIER']
                              ).toPandas()[self.config['INDIVIDUAL_IDENTIFIER']])
        self.data = self.data.withColumn(
            'val_flagged', self.data[self.config['INDIVIDUAL_IDENTIFIER']].isin(unique_list))
        return self.data

    def build_reserved_cols(self) -> pyspark.sql.DataFrame:
        """
        Add data split and outcome-related columns to the data.

        Returns:
            Spark DataFrame with reserved columns added
        """

        max_val = lit(self.data.agg({'_period': 'max'}).first()[0])
        self.data = self.data.withColumn(
            '_predict_obs', self.data['_period'] == max_val)

        max_test_intervals = int(
            (self.data.select('_period').distinct().count() - 1) / 2)
        if self.config.get('TEST_INTERVALS', 0) > max_test_intervals:
            warn(
                f"The specified value for TEST_INTERVALS was too high and will not allow for enough training periods. It was automatically reduced to {max_test_intervals}"
            )
            self.config['TEST_INTERVALS'] = max_test_intervals

        self.data = self.data.withColumn(
            '_test', (self.data['_period'] + self.config.get('TEST_INTERVALS', 0)) >= max_val)

        self.data = self.flag_validation_individuals()
        self.data = self.data.withColumn(
            '_validation', ~self.data['_test'] & self.data['val_flagged'])

        obs_max_window = Window.partitionBy(
            '_test').orderBy(col("_period").desc())
        self.data = self.data.withColumn('obs_max_period', F.max(
            self.data['_period']).over(obs_max_window))
        self.data = self.data.withColumn(
            '_maximum_lead', self.data.obs_max_period - self.data['_period'] + (
                self.data.obs_max_period < max_val).cast('int'))

        period_window = Window.partitionBy(
            self.config['INDIVIDUAL_IDENTIFIER']).orderBy('_period')
        self.data = self.data.withColumn(
            "gaps", lag(self.data['_period']).over(period_window) < (
                self.data['_period'] - 1)).fillna(False).withColumn(
                    "gaps", col('gaps').cast('int'))

        self.data = self.data.withColumn(
            "_spell", F.sum(self.data.gaps).over(period_window))

        duration_window = Window.partitionBy(
            [self.config['INDIVIDUAL_IDENTIFIER'], '_spell']).orderBy("_period")
        self.data = self.data.withColumn('_duration', F.count(
            self.data['_spell']).over(duration_window))

        max_duration_window = Window.partitionBy(
            [self.config['INDIVIDUAL_IDENTIFIER'], '_spell']).orderBy(col("_period").desc())
        self.data = self.data.withColumn('_duration', F.max(
            self.data['_duration']).over(max_duration_window) - self.data['_duration'])
        self.data = self.data.withColumn(
            '_event_observed', self.data['_duration'] < self.data['_maximum_lead'])

        self.data = self.data.drop('gaps')
        self.data = self.data.drop('val_flagged')
        self.data = self.data.drop('obs_max_period')
        return self.data

    def sort_panel_data(self) -> pyspark.sql.DataFrame:
        """
        Sort the data by individual, then by period.

        Returns:
            Sorted panel data
        """

        return self.data.orderBy(F.asc(self.config['INDIVIDUAL_IDENTIFIER']), F.asc('_period'))

    def build_processed_data(self):
        """
        Clean, augment, and store a panel dataset and related information.

        Returns:
            Processed data
        """

        if self.config.get('CACHE', False) == True:
            self.data.cache()
        self.check_panel_consistency()
        self.data = self.process_all_columns()
        self.data = self.build_reserved_cols()
        self.data = self.sort_panel_data()
        return self.data
